"""
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import hashlib
import json
import logging
import time
from datetime import datetime
from typing import Any, Optional
from urllib import parse

from urllib3 import PoolManager
from urllib3.exceptions import HTTPError

from sleet.constants import API_VALID_REGIONS, DEFAULT_CLIENT_HEADERS
from sleet.exceptions import APIStatusError, RegionError, SleetError
from sleet.oauth import OAuth

logger = logging.getLogger(__name__)

MAXIMUM_REQUESTS_HOUR = 36000
MAXIMUM_REQUESTS_SECOND = 100
NANOSECONDS_IN_HOUR = 3600000000000
NANOSECONDS_IN_SECOND = 1000000000


class RateLimit:
    last_reset_hourly = 0
    request_count_hourly = 0
    last_reset_seconds = 0
    request_count_seconds = 0

    @classmethod
    def check_rate_limits(cls) -> bool:
        now = time.monotonic_ns()

        # Has it been an hour since the last hourly request reset?
        if now - cls.last_reset_hourly >= NANOSECONDS_IN_HOUR:
            # Reset all limits
            cls.last_reset_hourly = now
            cls.request_count_hourly = 0

        # Has it been a second since the last second request reset?
        if now - cls.last_reset_seconds >= NANOSECONDS_IN_SECOND:
            # Reset all limits
            cls.last_reset_seconds = now
            cls.request_count_seconds = 0

        if cls.request_count_hourly >= MAXIMUM_REQUESTS_HOUR:
            return False

        if cls.request_count_seconds >= MAXIMUM_REQUESTS_SECOND:
            return False

        cls.request_count_seconds += 1
        cls.request_count_hourly += 1

        return True


class BaseAPI:
    hashing_enabled: bool = False
    rate_checking: bool = True

    def __init__(self, *, api_region: str, api_id: str, api_secret: str):
        self._api_region = api_region.lower()
        if self._api_region not in API_VALID_REGIONS:
            raise RegionError(self._api_region)

        self._base_api_url = f"https://{self._api_region}.api.blizzard.com".lower()
        self._client: PoolManager = PoolManager(headers=DEFAULT_CLIENT_HEADERS)
        self._oauth = OAuth(self._api_region, api_id, api_secret)

    def _request(
            self,
            resource: str,
            parameters: Optional[dict[str, Any]],
            headers: Optional[dict[str, str]]
    ) -> dict[str, Any]:
        """
        :param resource: The resource endpoint to be queried.
        :param parameters: Additional url parameters to be sent.
        :param headers: Addition request headers to be sent (eg. If-Modified-Since).
        :return: The JSON response from Blizzard.
        """

        if self.rate_checking:
            if not RateLimit.check_rate_limits():
                raise SleetError("Too many requests, slow down!")

        resource = f"{self._base_api_url}{resource}"
        if parameters:
            resource = f"{resource}?{parse.urlencode(parameters)}"

        headers = (headers or {}) | self._client.headers | {
            "Authorization": f"Bearer {self._oauth.token}",
            "X-Sleet-Requested-At": f"{datetime.utcnow().strftime('%a, %d %b %Y %H:%M:%S GMT')}"
        }

        logger.debug(f"Requesting {resource}. Headers: {headers}. Parameters: {parameters}.")

        try:
            response = self._client.request("GET", resource, headers=headers)

        except HTTPError as ex:
            raise SleetError(f"Error querying {resource}.") from ex

        if response.status != 200:
            raise APIStatusError(response.status)

        return self._response_unpack(resource, headers, response)

    def _response_unpack(self, url: str, headers: dict[str, str], response: Any) -> dict[str, Any]:
        """
        Decode an Urllib3 response.
        - A @metadata field will be created that contains the request/response headers.
        - The SHA256 hash of the response data (if hashing_enabled is set to True)

        :param url: The fully qualified URL the request was sent to
        :param headers: The additional headers sent in the request
        :param response: The response object from Urllib3
        :return: The API response with added metadata values
        """
        time_taken = (datetime.utcnow() -
                      datetime.strptime(headers["X-Sleet-Requested-At"], "%a, %d %b %Y %H:%M:%S GMT")).total_seconds()

        metadata = {
            "url": url,
            "headers": {
                "request": headers,
                "response": dict(response.headers)
            },
            "sha256": None,
            "elapsed": round(time_taken, 3)
        }

        if self.hashing_enabled:
            m = hashlib.sha256()
            m.update(response.data)
            metadata["sha256"] = m.hexdigest()

        response = response.data.decode()

        try:
            response = json.loads(response)

        except json.JSONDecodeError as ex:
            logger.error(f"Failed to obtain new token. Exception: {ex}.")
            raise SleetError(f"Failed to obtain new token. Exception: {ex}.")

        response["@metadata"] = metadata

        return response

    @staticmethod
    def _create_modified_since_header(modified_since: datetime) -> dict[str, str]:
        """
        Create an If-Modified-Since HTTP header from a Python datetime object.

        :param modified_since: A Python datetime object
        :return: A valid If-Modified-Since HTTP header
        """
        if not isinstance(modified_since, datetime):
            raise TypeError(f"Supplied modified_since '{modified_since}' is not a valid datetime.")

        return {"If-Modified-Since": modified_since.strftime("%a, %d %b %Y %H:%M:%S GMT")}
